function [m,output]=ec_fac_cv(theta,J,par,debug_draw)
% ec_fac_seqc.m 
% expectation consistent factorized inference
% for binary variables
% optimized using sequential variational bounding
% "guarenteed convergent" 

% Ole Winther, IMM, February 2005
  
% initialize variables
N = length(theta); 

% initialize constants
try draw = par.draw; catch draw = 0; end
try ftol=par.ftol; catch ftol = 1e-7; end  
try minchi=par.minchi; catch minchi = 1e-5; end
try m=par.m; catch m=zeros(N,1); end % initial value of m
try max_ite=par.max_ite; catch max_ite=100; end
try eta=par.eta; catch eta=1; end % set learning rate for lam_r update
try G = par.G; catch G = 0; end
try history = par.history; catch history = 0; end
try ro = par.ro; catch ro = 1; end % toggle random order 
try sloppystar = par.sloppystar; catch sloppystar = 0; end % toggle greedy update lam_s_star

if draw==1,fprintf('** Start ec_fac_cv ************************************ N=%d\n',N);tic;end

try Lam_r = par.Lam_r; 
catch
  % make sure that chi = inv( diag(Lam_r) - J ) is positive definite
  eigJ=eig(J); mineigJ = min(eigJ); maxeigJ = max(eigJ);
  Lam_r = 0*maxeigJ + 2 * ( maxeigJ - mineigJ ) * ones(N,1) ;
end
try chi = par.chi; catch chi =  inv( diag(Lam_r) - J ) ; end

dm = zeros(N,1); m_r = zeros(N,1); v_r = zeros(N,1); 
m_star=m;
dm_star = zeros(N,1); gam_s_star = zeros(N,1); Lam_s_star = zeros(N,1);

% initialize lam_s_star
Lam_s_star =  1 ./ ( 1 - m.^2 ); %1 ./ diag(chi)  ; % 1 ./ ( 1 - m.^2 );
gam_s_star = Lam_s_star .* m ;

% set mean vale of r-distribution to be the same as m - no set gam_r to
% zero
gam_r = zeros(N,1) ; % ( diag(Lam_r) - J ) * m - theta ; 
m_r = m ; 
v_r = diag( chi ) ;

dmdm=inf; dmdm_old=0; ite=0; overcount = 0;

if ro 

  while dmdm>ftol & ite<max_ite 
    
    ite=ite+1;
    indx=randperm(N);

    for sindx=1:N
    
      cindx=indx(sindx);

      % find mean and variance of r
      m_r(cindx) = chi(cindx,:) * ( gam_r + theta ) ;
      v_r(cindx) = chi(cindx,cindx);
          
      % find Lagrange parameters of s 
      Lam_s1 = 1 / max( minchi , v_r(cindx) ) ;
      gam_s1 = Lam_s1 .* m_r(cindx); 
       
      % update lam_q
      gam_q = gam_s1 - gam_r(cindx) ; 
      %Lam_q = Lam_s1 - Lam_r(cindx) ;

      % update moments of q distribution
      m(cindx) = tanh( gam_q );
      %v_q = 1 - m(cindx)^2 ; 
    
      dm(cindx) = m(cindx) - m_r(cindx) ;
    
      % initialize lam_q_star
      gam_q_star = gam_s_star(cindx) - gam_r(cindx) ; 

      % solve for lam_q_star
      gam_q_star = solve_fix_binary(gam_q_star+gam_s1,gam_q_star,ftol*1e-5); 
    
      % update moments of lam_q_star distribution
      m_star(cindx) = tanh( gam_q_star );
      v_q_star = 1 - m_star(cindx)^2 ;
    
      dm_star(cindx) = m_star(cindx) - m_r(cindx) ;

      % find Lagrange parameters of s and update lam_s_star
      Lam_s2 = 1 / max( minchi , v_q_star ) ;
      gam_s2 = Lam_s2 * m_star(cindx) ;

      % update lam_r
      dgam_r = eta * ( gam_s2 - gam_s1 ) ; gam_r(cindx) = dgam_r + gam_r(cindx) ;
      dLam_r = eta * ( Lam_s2 - Lam_s1 ) ; Lam_r(cindx) = dLam_r + Lam_r(cindx) ;

      % update chi using Sherman-Woodbury
      chiv = chi(:,cindx);
      kappa = - dLam_r / ( 1 + dLam_r * chi(cindx,cindx) );
      chi = chi + kappa * chiv * chiv'; 
      
    end % over variables - sindx 
    
    dmdm = dm' * dm ;
  
    dmdm_star = dm_star' * dm_star ;
   
    if dmdm_star < ftol | ( dmdm < dmdm_old & sloppystar )
      % update lam_s_star
      Lam_s_star = 1 ./ max( minchi , 1 - m_star.^2 ) ;
      gam_s_star = Lam_s_star .* m_star ;
    end
    
    dmdm_old = dmdm ;
    
    if history
      output.m(:,ite) = m;
      output.m_r(:,ite) = m_r;
      output.v_r(:,ite) = v_r;
    end
  
  end % while
  
else % order decided by least satisifed variable

  while dmdm>ftol & ite<max_ite
  
    ite=ite+1;
 
    for sindx=1:N
     
      % find Lagrange parameters of s 
      Lam_s1_all = 1 ./ max( minchi , v_r ) ;
      gam_s1_all = Lam_s1_all .* m_r ; 
      
      % update lam_q
      gam_q_all = gam_s1_all - gam_r ; 

      % update moments of q distribution
      m = tanh( gam_q_all ) ;

      % update variable with largest deviation m - m_r 
      dm = m - m_r ;
      [val,cindx]= max(abs(dm)) ;
    
      % find Lagrange parameters for current variabe
      gam_q = gam_q_all(cindx);
      gam_s1 = gam_s1_all(cindx);
      Lam_s1 = Lam_s1_all(cindx);
    
      % initialize lam_q_star
      gam_q_star = gam_s_star(cindx) - gam_r(cindx) ; 

      % solve for lam_q_star
      gam_q_star = solve_fix_binary(gam_q_star+gam_s1,gam_q_star,ftol*1e-3); 
    
      % update moments of lam_q_star distribution
      m_star(cindx) = tanh( gam_q_star );
      v_q_star = 1 - m_star(cindx)^2 ;
    
      dm_star(cindx) = m_star(cindx) - m_r(cindx) ;

      % find Lagrange parameters of s and update lam_s_star
      Lam_s2 = 1 / max( minchi , v_q_star ) ;
      gam_s2 = Lam_s2 * m_star(cindx) ;

      % update lam_r
      dgam_r = eta * ( gam_s2 - gam_s1 ) ; gam_r(cindx) = dgam_r + gam_r(cindx) ;
      dLam_r = eta * ( Lam_s2 - Lam_s1 ) ; Lam_r(cindx) = dLam_r + Lam_r(cindx) ;
     
      % update chi using Sherman-Woodbury
      chiv = chi(:,cindx);
      kappa = - dLam_r / ( 1 + dLam_r * chi(cindx,cindx) );
      chi = chi + kappa * chiv * chiv'; 
 
      % recalculate moments of r for all variables
      m_r = chiv * ( kappa * chiv' * (gam_r + theta) + dgam_r ) + m_r ;
      v_r = diag(chi);
    
    end % over variables - sindx 
  
    dmdm = dm' * dm ;
  
    dmdm_star = dm_star' * dm_star ;
   
    if dmdm_star < ftol | ( dmdm < dmdm_old & sloppystar )
       % update lam_s_star
      Lam_s_star = 1 ./ max( minchi , 1 - m_star.^2 ) ;
      gam_s_star = Lam_s_star .* m_star ;
    end
    
    if history
      output.m(:,ite) = m;
      output.m_r(:,ite) = m_r;
      output.v_r(:,ite) = v_r;
    end

  end % while
  
end % else

if ite == max_ite output.convergence=0; else output.convergence=1; end

if G % calculate free energy
  gam_q = m ./ max( minchi , 1 - m.^2 ) - gam_r ; % gam_s - gam_r 
  G_q = - sum( -log(2) + abs( gam_q ) + log( 1 + exp( -2 * abs( gam_q ) ) ) ) + m' * gam_q ;
  G_r = 0.5 * ( m' * ( gam_r + theta ) - logdet(chi) - sum(Lam_r) ) - m' * theta ;
  G_s = - 0.5 * ( sum( log( max( minchi, 1 - m .* m ) ) ) + N ) ;
  output.G = G_q + G_r - G_s ;
end

if nargout > 1
  output.chi = chi;  
  output.dmdm = dmdm;
  output.gam_q = m ./ max( minchi , 1 - m.^2 ) - gam_r ; % gam_s - gam_r 
  output.gam_r = gam_r ;
  output.ite = ite ;
  output.Lam_r = Lam_r ;
end
  
if draw==1,fprintf('** End of ec_fac_cv ****** time %4.3f sec********** ite =%3d\n\n',toc,ite);end

function [y,ite] = solve_fix_binary(x,y0,ftol) 
% gamma equation for sequential convergent algorithm 
% binary variables
%
% solve m(y) / ( 1 - m(y)^2 ) + y = x
% with m(y) = tanh(y)
% 
% using Newton:
%
% f(y) = m(y) / ( 1 - m(y)^2 ) + y - x
% f'(y) = 2 / ( 1 - m(y)^2 )
%
% dy = - eta * (f'(y))^-1 f(y) 

% Ole Winther, IMM, August 2004

eta = 1; max_ite = 100; 

try ftol; catch ftol=1e-12; end

y = y0;

m = tanh(y);
E = 0.5 * ( m + ( 1 - m.^2 ) .* ( y - x) );
ite = 0;
while sum(abs(E)) > ftol & ite < max_ite 
   ite = ite + 1;
   y = y - eta * E;
   m = tanh(y);
   E = 0.5 * ( m + ( 1 - m.^2 ) .* ( y - x) );
end

function y = logdet(A)
% log(det(A)) where A is positive-definite.
% This is faster and more stable than using log(det(A)).

U = chol(A);
y = 2*sum(log(diag(U)));
