function [m,output]=ec_tree_cv(theta,J,par,debug_draw)
% ec_tree_cv.m 
% expectation consistent tree structured inference
% for binary variables
% optimized using sequential variational bounding
% "guarenteed convergent" 
 
% doesn't calculate free energy 

% Ole Winther, IMM, March 2005
  
% initialize variables
N = length(theta); 

% initialize constants
try ftol=par.ftol; catch ftol = 1e-7; end  
try minchi=par.minchi; catch minchi = 1e-5; end
try m=par.m; catch m=zeros(N,1); end % initial value of m
try max_ite=par.max_ite; catch max_ite=100; end
try eta=par.eta; catch eta=1; end % set learning rate for lam_r update
try draw = par.draw; catch draw = 0; end
try G = par.G; catch G = 0; end
try history = par.history; catch history = 0; end
try treeop = par.treeop; catch treeop='J'; end
switch lower(treeop)
  case 'j'
    [nodes,ncon,CM] = find_tree(abs(J)); % use maximal spanning tree abs(J)
  case 'j_pairs'
    [nodes,ncon] = find_pairs(abs(J)); % use large abs(J) to set connectivity
  case 'random'
    [nodes,ncon,CM] = find_tree(rand(N,N)); % random graph
  case 'correlation'
    [nodes,ncon,CM] = find_tree(par.Ccorr); % use predefined correlation matrix
  otherwise
    disp('unknown treeop')  
end   

if draw==1,fprintf('** Start ec_tree_cv ************************************ N=%d\n',N);tic;end

try Lam_r = par.Lam_r; 
catch
  % make sure that chi = inv( diag(Lam_r) - J ) is positive definite
  eigJ=eig(J); mineigJ = min(eigJ); maxeigJ = max(eigJ);
  Lam_r = 0*maxeigJ + 10 * ( maxeigJ - mineigJ ) * ones(N,1) ;
end
Lam_r = sparse(1:N,1:N,Lam_r);
try chi_r = par.chi; catch chi_r =  inv( Lam_r - J ) ; end

% set mean vale of r-distribution be the same as m
gam_r = zeros(N,1) ; %( Lam_r - J ) * m - theta ;
m_r = m ;

% solve for lam_s_star
[gam_s_star,Lam_s_star] = lambda_solve_Gauss_tree(m,chi_r,nodes,ncon) ;

% set lam_q
gam_q = gam_s_star - gam_r ;
Lam_q = Lam_s_star - Lam_r ;

if 0
% another way to initialize
gam_q = theta;
Lam_q = CM .* J ;

[m,chi] = tree_inference_binary( gam_q , - Lam_q ) ;
[gam_s_star,Lam_s_star] = lambda_solve_Gauss_tree(m,chi,nodes,ncon) ; 

% set lam_r
gam_r = gam_s_star - gam_q ;
Lam_r = Lam_s_star - Lam_q ;

chi_r = inv( Lam_r - J ) ;
m_r = m ;
end

% set chi_q
%chi_q = chi_r ;

Nnodes = length(nodes);

dm = zeros(N,1); dm_star = zeros(N,1); mm = zeros(Nnodes,N); 

dmdm=inf; ite=0;
while dmdm>ftol & ite<max_ite 
  ite=ite+1;
  iperm = randperm(Nnodes);

  for i=iperm % sequential solution of inner loop in gamma and Lambda
      
    ip = nodes{i}.indx;
    
    % propagate evidence r -> q
    
    % find moments of r for variables ij
    m_r(ip) = chi_r(ip,:) * ( gam_r + theta ); 
    
    % set separator s1 - perhaps find better name
    Lam_s1 = inv( chi_r(ip,ip) ) ; 
    gam_s1 = Lam_s1 * m_r(ip) ;
      
    % propagate evidence q_ij -> r
    
    % set input argument to inner loop optimization  
    gam_x =  gam_s1 + gam_q(ip) ;
    Lam_x =  Lam_s1(1,2) + Lam_q(ip(1),ip(2)) ;

    % solve innner
    [m_q,chi_q,invchi_q,cite] = ...
       solve_fix_binary_structured_fb([ gam_x ; Lam_x ], gam_q , Lam_q , nodes{i} , ftol*1e-2) ;
    
    % set separators s2 - perhaps find better name
    gam_s2 = invchi_q * m_q ;
    Lam_s2 = invchi_q ;
             
    dm_star(ip) = m_q - m_r(ip) ;

    mm(i,ip) = m_q ;
    
    % update lam_r
    dgam_r = eta * ( gam_s2 - gam_s1 ) ; gam_r(ip) = dgam_r + gam_r(ip) ;
    dLam_r = eta * ( Lam_s2 - Lam_s1 ) ; Lam_r(ip,ip) = dLam_r + Lam_r(ip,ip) ;

    % update lam_q
    gam_q(ip) = gam_s_star(ip) - gam_r(ip) ;
    Lam_q(ip,ip) = Lam_s_star(ip,ip) - Lam_r(ip,ip) ; 
     
    % update chi_r using Sherman-Morrison
    %logdetchi_r = logdetchi_r - log( det( eye(2) + DeltaLam_r *chi_r(ip,ip) ) );  
    chi_r = chi_r - chi_r(:,ip) * dLam_r * inv( eye(2) + chi_r(ip,ip) * dLam_r) * chi_r(ip,:);
        
  end
  
  dmdm_star = dm_star' * dm_star ;
  
  if dmdm_star < ftol % * 0.1
          
    [gam_s_star,Lam_s_star] = lambda_solve_Gauss_tree(m_r,chi_r,nodes,ncon) ; 
  
    gam_q = gam_s_star - gam_r ;
    Lam_q = Lam_s_star - Lam_r ;
    
    [m,chi] = tree_inference_binary( gam_q , - Lam_q .* CM ) ;
    
    dm = m - m_r ;
    
    dmdm = dm' * dm ;

  end
  
end

if G % calculate free energy
  output.G = Ftree(m,chi_r,Lam_r-J,Lam_q,nodes,ncon) ; % inv(chi_r)=Lam_r-J;
end

if nargout > 1
  output.chi = chi_r;  
  output.dmdm = dmdm;
  output.gam_q = gam_q ;
  output.gam_r = gam_r ;
  output.Lam_r = Lam_r ;
  output.ite = ite ;
  output.mm = mm ; 
end

if draw==1,fprintf('** End of ec_tree_cv ****** time %4.3f sec********** ite =%3d\n\n',toc,ite);end

function [links,ncon,CM] = find_tree(C,ncon)
% constructs a tree from the 
% non-negative N*N correlation matrix C

N = size(C,1);

try ncon; catch ncon=N-1; end
pairs = 0;

[Y,I] = sort(C(:));
I = flipud(I);

indx = 1;

graph = (1:N)';

CM=sparse(N,N);

while pairs < ncon % we are looking for N-1 pairs

    i = fix( ( I(indx) - 1 ) / N ) + 1;
    j = rem( I(indx)-1 , N ) + 1; 

    if (graph(i) ~= graph(j)) % make link if nodes does not belong to the same sub-graph

        pairs = pairs + 1 ;

        links{pairs}.indx = [ i j ] ;
        
        graph(find(graph == graph(j))) = graph(i); % relabel nodes    

        CM(i,j) = 1;

    end
    
    indx = indx + 1;
    
end

% set adjacency matraix to be symmetric
CM = CM + CM';

% for each link find adjacent links - probably not optimal code
for i=1:ncon
   
  i1=links{i}.indx(1);
  i2=links{i}.indx(2);
  k1=0; k2=0;  links{i}.links1 = [];  links{i}.links2 = [] ;
  for j=[1:i-1 i+1:ncon]
    if any(links{j}.indx == i1)
      k1 = k1 + 1;
      links{i}.links1(k1) = j ;
    end
    if any(links{j}.indx == i2)
      k2 = k2 + 1;
      links{i}.links2(k2) = j ;
    end
  end 
end

% find order of transversal for all subtrees rooted in each link
for i=1:ncon
  pindx = 0 ; 
  CMred=CM; CMred( links{i}.indx(1) , links{i}.indx(2) ) = 0 ;  
  CMred( links{i}.indx(2) , links{i}.indx(1) ) = 0 ;
  for parent = links{i}.indx     
    pindx = pindx + 1 ; 
    k = 1 ; level = 0 ;  
    A = CMred(parent,:) ; indx = find( A ) ; B = CMred ; B(:,parent) = 0 ;  
    while ~isempty(indx)
   
      level = level + 1 ;
    
      [ii,jj,s] = find( sparse(parent,parent,1,N,N) * B  ) ; 
      tmp(pindx).nlinks(level) = length(s) ; 
      tmp(pindx).linkorder(k:k+tmp(pindx).nlinks(level)-1,:) = [ ii jj ] ;
      k = k + tmp(pindx).nlinks(level) ;
      parent = indx ;
  
      B(:,indx) = 0 ; % to avoid going backwards
  
      A = A * B ;
      indx = find(A) ;
  
    end
    tmp(pindx).levels = level ;
    if level tmp(pindx).linkorder = flipud(tmp(pindx).linkorder) ; end
  end % parents
  links{i}.levels1 = tmp(1).levels ;
  links{i}.levels2 = tmp(2).levels ;
  if links{i}.levels1 
    links{i}.lo1 = tmp(1).linkorder; 
    links{i}.nlinks1 = tmp(1).nlinks; 
  else links{i}.lo1 = [] ; links{i}.nlinks1 = [] ; end
  if links{i}.levels2
    links{i}.lo2 = tmp(2).linkorder; 
    links{i}.nlinks2 = tmp(2).nlinks; 
  else links{i}.lo2 = [] ; links{i}.nlinks2 = [] ; end
  clear tmp
end % ncon



%graph
ncon = (sum(CM,1))';    


if 0
k = 1 ;
parent = 1 ;
level = 0 ; 
A = CM(parent,:) ; indx = find( A ) ; B = CM ; B(:,parent) = 0 ; 
while ~isempty(indx)

  [i,j,s] = find( sparse(parent,parent,1,N,N) * B  )   ;
  nlinks = length(s) ; 
  linkorder(k:k+nlinks-1,:) = [ i j ] ;
  k = k + nlinks ;
  parent = indx ;
  
  B(:,indx) = 0 ; % to avoid going backwars
  level = level + 1;
  cl(level).indx = indx ; 
  
  A = A * B ;
  indx = find(A) ;
  
end
end  
  
plotgraph=0;
if plotgraph 
    gplot(CM,[-cos(2*pi*(0:N-1)'/N) sin(2*pi*(0:N-1)'/N)],'o-'), hold on
    plot(-cos(2*pi*(0:N-1)'/N),sin(2*pi*(0:N-1)'/N),'*'), hold off, drawnow
    bigfig(2)
end
plotgraphgrid=0;
if plotgraphgrid 
    sqrtN=sqrt(N); % assume integer
    gplot(CM,0.125+[mod((0:N-1)',sqrtN)/sqrtN floor((0:N-1)'/sqrtN)/sqrtN],'o-'), hold on
    plot(0.125+mod((0:N-1)',sqrtN)/sqrtN,0.125+floor((0:N-1)'/sqrtN)/sqrtN,'*'), 
    hold off, drawnow
    bigfig(2)
end

function [gam,Lam,G] = lambda_solve_Gauss_tree(m,chi,nodes,ncon,nindx)
%
% Solves for the Lagrange parameters for a Gaussian tree model 
% such that they are consistent with a given m and chi. 
% Uses the tree factorization of the free energy

N = size(chi,1);

if nargin < 5
    
  % single node contribution
  Lam = sparse( 1:N , 1:N ,( 1 - ncon ) ./ diag( chi ) );
  %G = - 0.5 * N - 0.5 * ( 1 - ncon )' * log( diag( chi ) ) ;

  Npair = length(nodes);
    
  for i=1:Npair

    ip = nodes{i}.indx;
    %detchi = det( chi( ip , ip ) ) ;
    %Lam(ip,ip) = Lam(ip,ip) + ...
    %  [ chi(ip(2),ip(2))  -chi(ip(1),ip(2)) ; ...
    %	-chi(ip(1),ip(2))   chi(ip(1),ip(1)) ] / detchi;
    Lam(ip,ip) = Lam(ip,ip) + inv(chi(ip,ip)) ;
    %G = G - 0.5 * logdet( chi(ip,ip) ) ; % log( detchi );    

  end
  
  gam = Lam * m ; 
  
else % only update terms containing links in nindx
      
  Lam = sparse(N,N) ; eye2 = eye(2) ;
    
  % update current node nindx and the nodes it is linked to
  for i=[ nodes{nindx}.links1 nodes{nindx}.links2 ] 
    ip = nodes{i}.indx ;   
    Lam(ip,ip) = Lam(ip,ip) + chi(ip,ip) \ eye(2) ; % inv(chi(ip,ip)) ;    
  end
  
  ip = nodes{nindx}.indx;
  Lam(ip,ip) = Lam(ip,ip) + chi(ip,ip) \ eye(2) + ... %inv(chi(ip,ip)) + ...
      diag ( ( 1 - ncon(ip) ) ./ diag( chi(ip,ip) ) );

  gam = Lam(ip,:) * m ;
  
  Lam = Lam(ip,ip) ;

  
  if 0 
  for i=1:Npair

    ip = nodes{i}.indx;
    
    if ip(1) == indx(1) | ip(1) == indx(2) | ip(2) == indx(1) | ip(2) == indx(2)
      nodes{i}.Lam = inv(chi(ip,ip)) ;
      Lam(ip,ip) = Lam(ip,ip) + nodes{i}.Lam ;
    end
    %G = G - 0.5 * logdet( chi(ip,ip) ) ; % log( detchi );    

  end

  gam = Lam(indx,:) * m ;
  
  Lam = Lam(indx,indx) ;

  end

end

function [m,chi_q,ichi_q,ite] = solve_fix_binary_structured_fb(x,gam_q,Lam_q,node,ftol)
% gamma equation for sequential convergent algorithm 
% binary variables
%

% does not quite run yet. Probably due to new sign on y(3)-term. Check!

global tanhJ thetag m_r N

N = size(Lam_q,1) ;
tanhJ = -tanh( Lam_q ) ;
thetag = gam_q ;
m_r = sparse(N,N) ; 

try ftol; catch ftol=1e-12; end
eta = 1; max_ite = 100;

E=inf*ones(3,1); H = zeros(3,3);

% optimize parameters for pair ij
i = node.indx(1); j = node.indx(2) ;
y(3,1) = Lam_q(i,j) ;
   
% calculate contribution from rest on the tree
% send messages to nodes i and j 
Lam_q(i,j)=0; Lam_q(j,i)=0; 
collect_messages2(i,Lam_q,node.levels1,node.nlinks1,node.lo1) ;
collect_messages2(j,Lam_q,node.levels2,node.nlinks2,node.lo2) ;
    
% calculate effective external field
theta_mij = sum( atanh( m_r(node.indx,:) ) , 2 ) ;

% solve for gam_q + thetam_ij 
y(1:2) = gam_q(node.indx) + theta_mij ;
x(1:2) = x(1:2) + theta_mij ;


ite = 0;
while sum(abs(E)) > ftol & ite < max_ite 

   ite = ite + 1;

   sh_a = sinh( y(1) + y(2) );                
   ch_a = cosh( y(1) + y(2) );                 
   sh_d = sinh( y(1) - y(2) ); 
   ch_d = cosh( y(1) - y(2) );
   
   eLam_p = exp( y(3) - abs(y(3)) ); eLam_m = exp( - y(3) - abs(y(3)) );
   Z = eLam_m * ch_a + eLam_p * ch_d ;
   m = [ eLam_m * sh_a + eLam_p * sh_d   eLam_m * sh_a - eLam_p * sh_d  ]' / Z ; 
   Mij = ( eLam_m * ch_a - eLam_p * ch_d ) / Z ;
   chi_q = [1 Mij; Mij 1] -  m * m' ;
   
   vi=chi_q(1,1); vj=chi_q(2,2); vij=chi_q(1,2);
   
   ichi_q =inv(chi_q);

   E_old = E;
   
   E(1:2) = ichi_q * m + ( y(1:2) - x(1:2) ) ;
   E(3) =  ichi_q(1,2) + ( y(3) - x(3) ) ;

   ichi_qm = ichi_q * m ;
   
   dmidy3 = Mij * m(1) - m(2); 
   offd = - dmidy3 - vi * m(2) - vij * m(1) ;
   dchi_qdy1 = [ - 2 * m(1) * vi offd ; offd ( - 2 * m(2) * vij ) ] ;
   H(1:2,1) =  - ichi_q * dchi_qdy1 * ichi_qm ;
   
   dmjdy3 = Mij * m(2) - m(1) ; 
   offd = - dmjdy3 - vj * m(1) - vij * m(2) ;
   dchi_qdy2 = [- 2 * m(1) * vij offd ; offd  ( - 2 * m(2) * vj ) ] ;
   H(1:2,2) = - ichi_q * dchi_qdy2 * ichi_qm ;
   
   H(1:2,1:2) = H(1:2,1:2) + eye(2);
   
   offd = Mij^2 - 1 - m(1) * dmjdy3 - m(2) * dmidy3 ;
   dchi_qdy3 = [ - 2 * m(1) * dmidy3 offd ; offd ( - 2 * m(2) * dmjdy3 ) ] ;
   H(3,3) = - ichi_q(1,:) * dchi_qdy3 * ichi_q(:,2) ;
   
   H(3,1) = - ichi_q(1,:) * dchi_qdy1 * ichi_q(:,2) ;
   H(3,2) = - ichi_q(1,:) * dchi_qdy2 * ichi_q(:,2) ;
   
   H(1:2,3) = ichi_q * ( - dchi_qdy3 * ichi_qm  + [ dmidy3 ; dmjdy3 ] ) ;
   
   H = H + eye(3) ;
   %H = diag(diag(H));
   
   y = y - eta * H \ E ;

end

y(1:2) = y(1:2) - theta_mij ;


function collect_messages2(i,CM,level,nlinks,linkorder)
% collect messages non recursively
global tanhJ thetag m_r N

% calculate messsages
n = 1; a = zeros(N,1) ;
for lindx=level:-1:1
   indx = n : n + nlinks(lindx) - 1 ;
   i = linkorder(indx,1) ; j = linkorder(indx,2) ; 
   % perhaps it is a good idea to use a sparse adjacency matrix instead of i and j 
   %m_q_new = sparse( i , j , tanh( thetag(j) + a(j) ) , N , N ) ;
   %m_q = m_q + m_q_new ;
   %m_r = m_r + m_q_new .* tanhJ  % message from factor (ij) to node i
   m_r_new = sparse( i , j , tanh( thetag(j) + a(j) ) , N , N ) .* tanhJ ; 
   m_r = m_r + m_r_new ;
   a = sum( atanh( m_r_new ) , 2 ) ;
   n = n + nlinks(lindx) ; 
end

function [m,chi] = tree_inference_binary(theta,J,node)
%
% exact inference on tree for binary variables
% using message-parsing

% for other multinomial variables use matrix multiplication. 

% possible optimizations: work in atanh_m_r and avoid calls when child is empty 

% apparently recursive calls makes this function very slow!
% reprogram as non-recursive.

% only collect messages so far reprogrammed a non-recursive.

% distribute_messages and calculate_chi should also be reprogrammed as
% nonrecursive .

% Ole Winther, IMM, March 2005

global tanhJ thetag m_r m_q chi m N

N=size(J,1); 
tanhJ = tanh( J ) ;
thetag = theta ;

%[i,j,s] = find(J) ;
m_r = sparse(N,N) ; %0*J ;
m_q = sparse(N,N) ; %0*J ;
  
if nargin < 3 
    
  % calculate m and chi for whole tree

  % find root node as arbritary leaf node
  CM = J ~= 0 ;
  ncon = sum(CM,2) ;
  i = find(ncon == 1) ;
  
  % send message to and from root
  collect_messages(i(1),CM) ;
  distribute_messages(i(1),0,CM) ;

  % calculate mean value 
  m = tanh( theta + sum(atanh(m_r),2) ) ;
 
  % calculate chi
  chi = sparse( 1:N , 1:N , 1 - m.^2 ) ;
  calculate_chi(i(1),0,J) ;

else

  %if length(node) == 2 
   
    % calculate pair correlation for node  
    % calculate m and chi for pair
    i = node.indx(1); j = node.indx(2) ;
    expJ = exp( J(i,j) - abs( J(i,j) ) ) ;
    expmJ = exp( - J(i,j) - abs( J(i,j) ) ) ;
    
    % send messages to node 
    J(i,j)=0; J(j,i)=0; 
    %collect_messages(i,J) ;
    %collect_messages(j,J) ;
    
    collect_messages2(i,J,node.levels1,node.nlinks1,node.lo1) ;
    collect_messages2(j,J,node.levels2,node.nlinks2,node.lo2) ;  
    
    % calculate mean value and chi    
    theta = theta(node.indx) + sum( atanh( m_r(node.indx,:) ) , 2 ) ;
    ca = cosh( theta(1) + theta(2) ) ; cm = cosh( theta(1) - theta(2) ) ;
    sa = sinh( theta(1) + theta(2) ) ; sm = sinh( theta(1) - theta(2) ) ;
    Z = expJ * ca + expmJ * cm ;
    m = [ expJ * sa + expmJ * sm   expJ * sa - expmJ * sm  ]' / Z ;
    Mij = ( expJ * ca - expmJ * cm ) / Z;
    chi = [1 Mij; Mij 1] -  m * m'; 

  %else
  if 0 
  
    % calculate m and chi for single variable
    % node is assumed to be scalar vector
  
    % send messages to node  
    collect_messages(node(1),J) ;

    % calculate mean value 
    m = tanh( theta(node(1)) + sum( atanh( m_r(node(1),:) ) ) );
    
    % calculate chi
    chi = 1 - m.^2 ;
    
  end
end

function m_q_cur = collect_messages(i,CM)
% collect messages
global tanhJ thetag m_r m_q

%i
%CM(i,:)
child_i = find(CM(i,:)) ;

CM(child_i,i) = 0 ; % to avoid going back

a = thetag(i);

for j = child_i  
  m_q(i,j) = collect_messages(j,CM) ; % message from node j to factor (ij)
  m_r(i,j) = m_q(i,j) * tanhJ(i,j) ; % message from factor (ij) to node i
  a = a + atanh( m_r(i,j) ) ; % 0.5 * log( ( 1 + m_r(i,j) ) / ( 1 - m_r(i,j) ) ) ;
end

m_q_cur = tanh( a ) ;

function distribute_messages(i,m_r_cur,CM)
% distribute messages
global tanhJ thetag m_r m_q chi

%i
%CM(i,:)
child_i = find(CM(i,:)) ;

CM(child_i,i) = 0 ; % to avoid going back

a = thetag(i);

a_child = atanh( m_r(i,child_i) ) ; 

a = a + atanh(m_r_cur) + sum( a_child ) ; 

k = 0 ;
for j = child_i
  k = k + 1 ;
  m_q(j,i) = tanh( a - a_child(k) ) ; % message from node i to factor (ij)  
  m_r(j,i) = m_q(j,i) * tanhJ(j,i) ; % message from factor (ij) to node j
  distribute_messages(j,m_r(j,i),CM) ;
end

function calculate_chi(i,m_r_cur,J)
% calculates chi going through the tree resursively
global tanhJ thetag m_r m_q chi m 

%i
child_i = find(J(i,:)) ;

expJ = exp( J(child_i,i) - abs( J(child_i,i) ) ) ;
expmJ = exp( - J(child_i,i) - abs( J(child_i,i) ) ) ;
J(child_i,i) = 0 ; % to avoid going back

a_child = atanh( m_r(i,child_i) ) ;

theta = thetag(i) + sum( a_child ) + atanh( m_r_cur ) ; 

k = 0 ;
for j = child_i
  k = k + 1 ;
  theta_i = theta - a_child(k) ;  
  child_j = find( J(j,:) ) ;
  theta_j = thetag(j) + sum( atanh( m_r(j,child_j) ) ) ;
  ca = cosh( theta_i + theta_j ) ; cm = cosh( theta_i - theta_j ) ;
  chi(i,j) = ( ca * expJ(k) - cm * expmJ(k) ) / ( ca * expJ(k) + cm * expmJ(k) ) - m(i) * m(j) ;
  chi(j,i) = chi(i,j) ;
  calculate_chi(j,m_r(j,i),J) ;
end

function [F] = Ftree(m,chi_r,invchi_r,Lam_q,nodes,ncon)
% this function is only valid at convergence
% to look at each term independently one must define logZ_q and logZ_s in
% terms of their Lagrange parameters

N = length(m); 

logZ_r = 0.5 * ( m' * invchi_r * m + logdet(chi_r) ) ;

% single node contribution
logZ_q = ( 1 - ncon )' * log( 2*cosh( atanh( m ) ) ) - 0.5 * sum( diag( Lam_q ) ) ; 
logZ_s = 0.5 * ( 1 - ncon )' * ( log( diag( chi_r ) ) + m.^2 ./ diag( chi_r ) ) ;

Npair = length(nodes);
    
for i=1:Npair

  ip = nodes{i}.indx ;
 
  % update logZ_s
  logZ_s = logZ_s + 0.5 * logdet( chi_r(ip,ip) ) + 0.5 * m(ip)' * inv( chi_r(ip,ip) ) * m(ip) ; 
  
  % update logZ_q
  m1 = m(ip(1)) ; m2 = m(ip(2)) ;
  M = chi_r(ip(1),ip(2)) + m1 * m2;
  gam1 = 0.5 * atanh( ( m1 + m2 ) / ( 1 + M ) ) + 0.5 * atanh( ( m1 - m2 ) / ( 1 - M ) ) ;
  gam2 = 0.5 * atanh( ( m1 + m2 ) / ( 1 + M ) ) + 0.5 * atanh( ( m2 - m1) / ( 1 - M ) ) ;
  Lam_qij =  - 0.5 * atanh( ( m1 + M ) / ( 1 + m2 ) ) - 0.5 * atanh( ( M - m1 ) / ( 1 - m2 ) ) ;
  
  logZ_q = logZ_q + log( 2*cosh(gam1+gam2)*exp(-Lam_qij)+ 2*cosh(gam1-gam2)*exp(Lam_qij) ) ;
  
end

F = -logZ_q - logZ_r + logZ_s ;

function y = logdet(A)
% log(det(A)) where A is positive-definite.
% This is faster and more stable than using log(det(A)).

U = chol(A);
y = 2*sum(log(diag(U)));