function [links,ncon,CM] = find_tree(C,ncon)
% constructs a tree from the 
% non-negative N*N correlation matrix C

N = size(C,1);

try ncon; catch ncon=N-1; end
pairs = 0;

[Y,I] = sort(C(:));
I = flipud(I);

indx = 1;

graph = (1:N)';

CM=sparse(N,N);

while pairs < ncon % we are looking for N-1 pairs

    i = fix( ( I(indx) - 1 ) / N ) + 1;
    j = rem( I(indx)-1 , N ) + 1; 

    if (graph(i) ~= graph(j)) % make link if nodes does not belong to the same sub-graph

        pairs = pairs + 1 ;

        links{pairs}.indx = [ i j ] ;
        
        graph(find(graph == graph(j))) = graph(i); % relabel nodes    

        CM(i,j) = 1;

    end
    
    indx = indx + 1;
    
end

% set adjacency matraix to be symmetric
CM = CM + CM';

% for each link find adjacent links - probably not optimal code
for i=1:ncon
   
  i1=links{i}.indx(1);
  i2=links{i}.indx(2);
  k1=0; k2=0;  links{i}.links1 = [];  links{i}.links2 = [] ;
  for j=[1:i-1 i+1:ncon]
    if any(links{j}.indx == i1)
      k1 = k1 + 1;
      links{i}.links1(k1) = j ;
    end
    if any(links{j}.indx == i2)
      k2 = k2 + 1;
      links{i}.links2(k2) = j ;
    end
  end 
end

% find order of transversal for all subtrees rooted in each link
for i=1:ncon
  pindx = 0 ; 
  CMred=CM; CMred( links{i}.indx(1) , links{i}.indx(2) ) = 0 ;  
  CMred( links{i}.indx(2) , links{i}.indx(1) ) = 0 ;
  for parent = links{i}.indx     
    pindx = pindx + 1 ; 
    k = 1 ; level = 0 ;  
    A = CMred(parent,:) ; indx = find( A ) ; B = CMred ; B(:,parent) = 0 ;  
    while ~isempty(indx)
   
      level = level + 1 ;
    
      [ii,jj,s] = find( sparse(parent,parent,1,N,N) * B  ) ; 
      tmp(pindx).nlinks(level) = length(s) ; 
      tmp(pindx).linkorder(k:k+tmp(pindx).nlinks(level)-1,:) = [ ii jj ] ;
      k = k + tmp(pindx).nlinks(level) ;
      parent = indx ;
  
      B(:,indx) = 0 ; % to avoid going backwards
  
      A = A * B ;
      indx = find(A) ;
  
    end
    tmp(pindx).levels = level ;
    if level tmp(pindx).linkorder = flipud(tmp(pindx).linkorder) ; end
  end % parents
  links{i}.levels1 = tmp(1).levels ;
  links{i}.levels2 = tmp(2).levels ;
  if links{i}.levels1 
    links{i}.lo1 = tmp(1).linkorder; 
    links{i}.nlinks1 = tmp(1).nlinks; 
  else links{i}.lo1 = [] ; links{i}.nlinks1 = [] ; end
  if links{i}.levels2
    links{i}.lo2 = tmp(2).linkorder; 
    links{i}.nlinks2 = tmp(2).nlinks; 
  else links{i}.lo2 = [] ; links{i}.nlinks2 = [] ; end
  clear tmp
end % ncon



%graph
ncon = (sum(CM,1))';    


if 0
k = 1 ;
parent = 1 ;
level = 0 ; 
A = CM(parent,:) ; indx = find( A ) ; B = CM ; B(:,parent) = 0 ; 
while ~isempty(indx)

  [i,j,s] = find( sparse(parent,parent,1,N,N) * B  )   ;
  nlinks = length(s) ; 
  linkorder(k:k+nlinks-1,:) = [ i j ] ;
  k = k + nlinks ;
  parent = indx ;
  
  B(:,indx) = 0 ; % to avoid going backwars
  level = level + 1;
  cl(level).indx = indx ; 
  
  A = A * B ;
  indx = find(A) ;
  
end
end  
  
plotgraph=0;
if plotgraph 
    gplot(CM,[-cos(2*pi*(0:N-1)'/N) sin(2*pi*(0:N-1)'/N)],'o-'), hold on
    plot(-cos(2*pi*(0:N-1)'/N),sin(2*pi*(0:N-1)'/N),'*'), hold off, drawnow
    bigfig(2)
end
plotgraphgrid=0;
if plotgraphgrid 
    sqrtN=sqrt(N); % assume integer
    gplot(CM,0.125+[mod((0:N-1)',sqrtN)/sqrtN floor((0:N-1)'/sqrtN)/sqrtN],'o-'), hold on
    plot(0.125+mod((0:N-1)',sqrtN)/sqrtN,0.125+floor((0:N-1)'/sqrtN)/sqrtN,'*'), 
    hold off, drawnow
    bigfig(2)
end