trials = 1;
save_results=0;
display_summary_results = 0 ;

% set up problem 
dobs = 0.25 ; 
dcoup = 0.5 ;
N = 16; %19*19;
interaction = 'mixed';
grid = 1;

% toggle which algorithms to run
run_ec_fac_cv = 0; % sequential convergent
run_ec_tree_cv = 0; % sequential convergent
run_ec_fac_ep = 0; % expectation propagation style
run_ec_tree_ep = 0; % tree expectation propagation style
run_ec_fac_jtree = 0; % fac ec parallel update
run_ec_tree_jtree = 0; % tree ec parallel update by Wim Wiegerinck
run_ec_fac_seqc = 0; % "greedy sequential convergent" 
run_ec_tree_seqc = 0; % tree "greedy sequential convergent"
run_ec_fac_prag = 1; % pragmatic approach first single loop then d-loop
run_ec_tree_prag = 1; % pragmatic approach first single loop then d-loop


% set algorithm options
par.max_ite=500; 
par.ftol=1e-15;
par.draw=1;
par.G = 0;
par.minchi=1e-8;
par.sloppystar = 0; % greedy update of lam_s_star in ec_fac_cv, makes it somewhat faster
%par.treeop = 'j_pairs';

% addtional variables
mean_ite_cv = 0 ;
mean_ite_tcv = 0 ;
mean_ite_seqc = 0 ;
mean_ite_tseqc = 0 ;
mean_ite_ep = 0 ;
mean_ite_tep = 0 ;
mean_ite_jtree = 0;
mean_ite_tjtree = 0 ;
mean_ite_prag = 0 ;
mean_ite_tprag = 0 ;
ep_converged = 0;
tep_converged = 0;
jtree_converged = 0;
tjtree_converged = 0;

if grid % only valid for sqaure grid
  mask = mask_funk(N);
end

for t=1:trials
    
  seeds(t) = t; %rand('seed');
  seedsn(t) = t; %randn('seed');

  rand('seed',seeds(t)); randn('seed',seedsn(t));

  theta = 2 * dobs * rand(N,1) - dobs; %  external field
  J = rand(N,N); J = triu(J,1); J = J + J';
  
  switch lower(interaction)
    case 'repulsive'   % (a) repulsive (anti-ferromagnetic)
      J = - 2 * dcoup * J;                              
    case 'mixed'       % (b) mixed (paramagnetic)
      J = 2 * dcoup * J - dcoup; J = J - diag(diag(J)); 
    case 'attractive'  % (c) attractive (ferromagnetic)
      J = 2 * dcoup * J;
    otherwise
      disp('unknown interaction type')  
  end

  if grid
    J = mask .* J;
    [i,j,s] = find(J);
    J = sparse(i,j,s);
    clear i j s
  end

  [mexact,chiexact,Gexact]=exact(theta,J) ;
   
  if run_ec_fac_cv
    %par.G = 1 ; % calculate free energy
    [mcv,output]=ec_fac_cv(theta,J,par);
    Error_cv(t) = 0.5 * sum(abs(mcv-mexact)) / N ;
    E_cv = Error_cv(t)
    mean_ite_cv = mean_ite_cv + output.ite ;
    %G=output.G
  end
 
  if run_ec_tree_cv
    [mtcv,output]=ec_tree_cv(theta,J,par);
    Error_tcv(t) = 0.5 * sum(abs(mtcv-mexact)) / N ;
    E_tcv = Error_tcv(t)
    mean_ite_tcv = mean_ite_tcv + output.ite ;
  end
  
  if run_ec_fac_ep
    [mep,output]=ec_fac_ep(theta,J,par);
    Error_ep(t) = 0.5 * sum(abs(mep-mexact)) / N ; 
    E_ep = Error_ep(t)
    ep_converged=ep_converged+output.converged;
    mean_ite_ep = mean_ite_ep + output.ite ;
  end

  if run_ec_tree_ep
    %par.G = 1 ; % calculate free energy
    [mtep,output]=ec_tree_ep(theta,J,par);
    Error_tep(t) = 0.5 * sum(abs(mtep-mexact)) / N ;
    E_tep = Error_tep(t)
    tep_converged=tep_converged+output.converged;
    mean_ite_tep = mean_ite_tep + output.ite ;
    %G = output.G
  end
  
  if run_ec_fac_jtree
    [mjtree,output]=ec_fac_jtree(theta,J,par);
    Error_jtree(t) = 0.5 * sum(abs(mjtree-mexact)) / N ;
    E_jtree = Error_jtree(t)
    jtree_converged=jtree_converged+output.converged;
    mean_ite_jtree = mean_ite_jtree + output.ite ;
  end

  if run_ec_tree_jtree
    %par.G = 1 ; % calculate free energy
    [mtjtree,output]=ec_tree_jtree(theta,J,par);
    Error_tjtree(t) = 0.5 * sum(abs(mtjtree-mexact)) / N ;
    E_tjtree = Error_tjtree(t)
    tjtree_converged=tjtree_converged+output.converged;
    mean_ite_tjtree = mean_ite_tjtree + output.ite ;
    %G = output.G
  end
  
  if run_ec_fac_seqc
    [mseqc,output]=ec_fac_jtree(theta,J,par);
    Error_seqc(t) = 0.5 * sum(abs(mseqc-mexact)) / N ;  
    E_seqc = Error_seqc(t)
    mean_ite_seqc = mean_ite_seqc + output.ite ;
  end

  if run_ec_tree_seqc
    par.G = 1 ; % calculate free energy
    [mtseqc,output]=ec_tree_seqc(theta,J,par);
    Error_tseqc(t) = 0.5 * sum(abs(mtseqc-mexact)) / N ;
    E_tseqc = Error_tseqc(t)
    mean_ite_tseqc = mean_ite_tseqc + output.ite ;
    G = output.G
  end
  
  if run_ec_fac_prag
    [mprag,output]=ec_fac_ep(theta,J,par);
    mean_ite_prag = mean_ite_prag + output.ite ;
    if ~output.converged
      [mprag,output]=ec_fac_cv(theta,J,par);
      mean_ite_prag = mean_ite_prag + output.ite ;
    end
    Error_prag(t) = 0.5 * sum(abs(mprag-mexact)) / N ;
    E_prag = Error_prag(t)
  end
  
  if run_ec_tree_prag
    [mtprag,output]=ec_tree_jtree(theta,J,par);
    mean_ite_tprag = mean_ite_tprag + output.ite ;
    if ~output.converged
      [mtprag,output]=ec_tree_cv(theta,J,par);
      mean_ite_tprag = mean_ite_tprag + output.ite ;
    end
    Error_tprag(t) = 0.5 * sum(abs(mtprag-mexact)) / N ;
    E_tprag = Error_tprag(t)
  end

end  


if display_summary_results
    
if run_ec_fac_cv
  Error_cv
  meanError_cv = mean(Error_cv)
  stdError_cv = std(Error_cv)
  medianError_cv = median(Error_cv)
  minError_cv = min(Error_cv)
  maxError_cv = max(Error_cv)
  mean_ite_cv = mean_ite_cv / trials
end

if run_ec_tree_cv
  Error_tcv
  meanError_tcv = mean(Error_tcv)
  stdError_tcv = std(Error_tcv)
  medianError_tcv = median(Error_tcv)
  minError_tcv = min(Error_tcv)
  maxError_tcv = max(Error_tcv)
  mean_ite_tcv = mean_ite_tcv / trials
end

if run_ec_fac_ep
  Error_ep
  meanError_ep = mean(Error_ep)
  stdError_ep = std(Error_ep)
  medianError_ep = median(Error_ep)
  minError_ep = min(Error_ep)
  maxError_ep = max(Error_ep)
  ec_converged = ec_converged / trials
  mean_ite_ep = mean_ite_ep / trials
end
  
if run_ec_tree_ep
  Error_tep
  meanError_tep = mean(Error_tep)
  stdError_tep = std(Error_tep)
  medianError_tep = median(Error_tep)
  minError_tep = min(Error_tep)
  maxError_tep = max(Error_tep)
  tep_converged = tep_converged / trials
  mean_ite_tep = mean_ite_tep / trials
end

if run_ec_fac_jtree
  Error_jtree
  meanError_jtree = mean(Error_jtree)
  stdError_jtree = std(Error_jtree)
  medianError_jtree = median(Error_jtree)
  minError_jtree = min(Error_jtree)
  maxError_jtree = max(Error_jtree)
  jtree_converged = jtree_converged / trials
  mean_ite_jtree = mean_ite_jtree / trials
end

if run_ec_tree_jtree
  Error_tjtree
  meanError_tjtree = mean(Error_tjtree)
  stdError_tjtree = std(Error_tjtree)
  medianError_tjtree = median(Error_tjtree)
  minError_tjtree = min(Error_tjtree)
  maxError_tjtree = max(Error_tjtree)
  tjtree_converged = tjtree_converged / trials
  mean_ite_tjtree = mean_ite_tjtree / trials
end

if run_ec_fac_seqc
  Error_seqc  
  meanError_seqc = mean(Error_seqc)
  stdError_seqc = std(Error_seqc)
  medianError_seqc = median(Error_seqc)
  minError_seqc = min(Error_seqc)
  maxError_seqc = max(Error_seqc)
  mean_ite_seqc = mean_ite_seqc / trials
end

if run_ec_tree_seqc
  Error_tseqc  
  meanError_tseqc = mean(Error_tseqc)
  stdError_tseqc = std(Error_tseqc)
  medianError_tseqc = median(Error_tseqc)
  minError_tseqc = min(Error_tseqc)
  maxError_tseqc = max(Error_tseqc)
  mean_ite_tseqc = mean_ite_tseqc / trials
end

if run_ec_fac_prag
  Error_prag  
  meanError_prag = mean(Error_prag)
  stdError_prag = std(Error_prag)
  medianError_prag = median(Error_prag)
  minError_prag = min(Error_prag)
  maxError_prag = max(Error_prag)
  mean_ite_prag = mean_ite_prag / trials
end

if run_ec_tree_prag
  Error_tprag  
  meanError_tprag = mean(Error_tprag)
  stdError_tprag = std(Error_tprag)
  medianError_tprag = median(Error_tprag)
  minError_tprag = min(Error_tprag)
  maxError_tprag = max(Error_tprag)
  mean_ite_tprag = mean_ite_tprag / trials
end

end

if save_results
  if grid
    s = sprintf('save wainwright_grid_%s_%g_%g.mat',interaction,dobs,dcoup);
  else % fully connected
    s = sprintf('save wainwright_fully_connected_%s_%g_%g.mat',interaction,dobs,dcoup);
  end
  eval(s)
end
